#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

from neptune.core.components.abstract import Resource
from neptune.internal.utils.logger import get_logger

logger = get_logger()


class LogFile(Resource):
    def __init__(self, data_path: Path, min_version: int, extension: str = "log") -> None:
        self._data_path: Path = data_path
        self._min_version: int = min_version
        self._extension: str = extension

        self._file_size: int = 0
        if (data_path / f"data-{min_version}.{extension}").exists():
            self._file_size = self.file_path.stat().st_size

        self._writer = open(self.file_path, "a")

    @property
    def data_path(self) -> Path:
        return self._data_path

    @property
    def min_version(self) -> int:
        return self._min_version

    @property
    def file_size(self) -> int:
        return self._file_size

    @property
    def file_name(self) -> str:
        return f"data-{self._min_version}.{self._extension}"

    @property
    def file_path(self) -> Path:
        return self._data_path / self.file_name

    def write(self, data: str) -> None:
        self._writer.write(data + "\n")
        self._file_size += len(data) + 1

    def cleanup(self) -> None:
        self.close()
        try:
            self.file_path.unlink()
        except FileNotFoundError:
            pass
        except Exception:
            logger.exception("Cannot remove queue file %s", self.file_name)

    def flush(self) -> None:
        if not self._writer.closed:
            self._writer.flush()

    def close(self) -> None:
        if not self._writer.closed:
            self._writer.close()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["JsonFileSplitter"]

from collections import deque
from io import StringIO
from json import (
    JSONDecodeError,
    JSONDecoder,
)
from pathlib import Path
from types import TracebackType
from typing import (
    IO,
    Any,
    Deque,
    Optional,
    Tuple,
    Type,
    Union,
)


class JsonFileSplitter:
    BUFFER_SIZE = 64 * 1024
    MAX_PART_READ = 8 * 1024

    def __init__(self, file_path: Union[str, Path]):
        self._file: IO = open(file_path, "r")
        self._decoder: JSONDecoder = JSONDecoder(strict=False)
        self._part_buffer: StringIO = StringIO()
        self._parsed_queue: Deque[Tuple[Any, int]] = deque()
        self._start_pos: int = 0

    def close(self) -> None:
        if not self._file.closed:
            self._file.close()
        if not self._part_buffer.closed:
            self._part_buffer.close()

    def get(self) -> Optional[dict]:
        return (self.get_with_size() or (None, None))[0]

    def get_with_size(self) -> Tuple[Optional[dict], int]:
        if self._parsed_queue:
            return self._parsed_queue.popleft()
        self._read_data()
        if self._parsed_queue:
            return self._parsed_queue.popleft()
        return None, 0

    def _read_data(self) -> None:
        if self._part_buffer.tell() < self.MAX_PART_READ:
            data = self._file.read(self.BUFFER_SIZE)
            if not data:
                return
            if self._part_buffer.tell() > 0:
                data = self._reset_part_buffer() + data
            self._decode(data)

        if not self._parsed_queue:
            data = self._file.read(self.BUFFER_SIZE)
            while data:
                self._part_buffer.write(data)
                data = self._file.read(self.BUFFER_SIZE)
            data = self._reset_part_buffer()
            self._decode(data)

    def _decode(self, data: str) -> None:
        start = self._json_start(data)
        while start is not None:
            try:
                json_data, new_start = self._decoder.raw_decode(data, start)
                size = new_start - start
                start = new_start
            except JSONDecodeError:
                self._part_buffer.write(data[start:])
                break
            else:
                self._parsed_queue.append((json_data, size))
                start = self._json_start(data, start)

    @staticmethod
    def _json_start(data: str, start: int = 0) -> Optional[int]:
        try:
            return data.index("{", start)
        except ValueError:
            return None

    def _reset_part_buffer(self) -> str:
        data = self._part_buffer.getvalue()
        self._part_buffer.close()
        self._part_buffer = StringIO()
        return data

    def __enter__(self) -> "JsonFileSplitter":
        return self

    def __exit__(
        self,
        exc_type: Type[Optional[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_logger", "get_disabled_logger", "NEPTUNE_LOGGER_NAME"]

import logging
import sys

NEPTUNE_LOGGER_NAME = "neptune"
NEPTUNE_NO_PREFIX_LOGGER_NAME = "neptune_no_prefix"
NEPTUNE_NOOP_LOGGER_NAME = "neptune_noop"
LOG_FORMAT = "[%(name)s] [%(levelname)s] %(message)s"
NO_PREFIX_FORMAT = "%(message)s"


class CustomFormatter(logging.Formatter):
    def format(self, record):
        record.levelname = record.levelname.lower().ljust(len("warning"))
        formatter = logging.Formatter(LOG_FORMAT)
        return formatter.format(record)


class GrabbableStdoutHandler(logging.StreamHandler):
    """
    This class is like a StreamHandler using sys.stdout, but always uses
    whatever sys.stdout is currently set to rather than the value of
    sys.stderr at handler construction time.
    This enables neptune-client to capture stdout regardless
    of logging configuration time.
    Based on logging._StderrHandler from standard library.
    """

    def __init__(self, level=logging.NOTSET):
        logging.Handler.__init__(self, level)

    @property
    def stream(self):
        return sys.stdout


def get_logger(with_prefix: bool = True) -> logging.Logger:
    name = NEPTUNE_LOGGER_NAME if with_prefix else NEPTUNE_NO_PREFIX_LOGGER_NAME

    return logging.getLogger(name)


def get_disabled_logger() -> logging.Logger:
    return logging.getLogger(NEPTUNE_NOOP_LOGGER_NAME)


def _set_up_logging():
    # setup neptune logger so that logging.getLogger(NEPTUNE_LOGGER_NAME)
    # returns configured logger
    neptune_logger = logging.getLogger(NEPTUNE_LOGGER_NAME)
    neptune_logger.propagate = False

    stdout_handler = GrabbableStdoutHandler()
    stdout_handler.setFormatter(CustomFormatter())
    neptune_logger.addHandler(stdout_handler)

    neptune_logger.setLevel(logging.INFO)


def _set_up_no_prefix_logging():
    neptune_logger = logging.getLogger(NEPTUNE_NO_PREFIX_LOGGER_NAME)
    neptune_logger.propagate = False

    stdout_handler = GrabbableStdoutHandler()
    stdout_handler.setFormatter(logging.Formatter(NO_PREFIX_FORMAT))
    neptune_logger.addHandler(stdout_handler)

    neptune_logger.setLevel(logging.INFO)


def _set_up_disabled_logging():
    neptune_logger = logging.getLogger(NEPTUNE_NOOP_LOGGER_NAME)

    neptune_logger.setLevel(logging.CRITICAL)


_set_up_logging()
_set_up_no_prefix_logging()
_set_up_disabled_logging()



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["CallbacksMonitor"]

from queue import Queue
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
)

from neptune.internal.background_job import BackgroundJob
from neptune.internal.signals_processing.signals_processor import SignalsProcessor

if TYPE_CHECKING:
    from neptune.internal.signals_processing.signals import Signal
    from neptune.metadata_containers import MetadataContainer


class CallbacksMonitor(BackgroundJob):
    def __init__(
        self,
        queue: "Queue[Signal]",
        async_lag_threshold: float,
        async_no_progress_threshold: float,
        async_lag_callback: Optional[Callable[["MetadataContainer"], None]] = None,
        async_no_progress_callback: Optional[Callable[["MetadataContainer"], None]] = None,
        period: float = 10,
    ) -> None:
        self._period: float = period
        self._queue: "Queue[Signal]" = queue
        self._thread: Optional["SignalsProcessor"] = None
        self._started: bool = False
        self._async_lag_threshold: float = async_lag_threshold
        self._async_no_progress_threshold: float = async_no_progress_threshold
        self._async_lag_callback: Optional[Callable[["MetadataContainer"], None]] = async_lag_callback
        self._async_no_progress_callback: Optional[Callable[["MetadataContainer"], None]] = async_no_progress_callback

    def start(self, container: "MetadataContainer") -> None:
        self._thread = SignalsProcessor(
            period=self._period,
            container=container,
            queue=self._queue,
            async_lag_threshold=self._async_lag_threshold,
            async_no_progress_threshold=self._async_no_progress_threshold,
            async_lag_callback=self._async_lag_callback,
            async_no_progress_callback=self._async_no_progress_callback,
        )
        self._thread.start()
        self._started = True

    def stop(self) -> None:
        if self._thread and self._started:
            self._thread.interrupt()

    def join(self, seconds: Optional[float] = None) -> None:
        if self._thread and self._started:
            self._thread.join(seconds)

    def pause(self) -> None:
        if self._thread:
            self._thread.pause()

    def resume(self) -> None:
        if self._thread:
            self._thread.resume()
